// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include "umfpack.h"
#include "sim.h"
#include "complex_solver.h"
#include <math.h>

static int last_col=0;
static int last_nz=0;
static double *x=NULL;
static double *xz=NULL;
static int *Ap=NULL;
static int *Ai=NULL;
static double *Ax=NULL;
static double *Az=NULL;

void complex_error_report(int status, const char *file, const char *func, int line)
{
	fprintf(stderr, "in %s: file %s, line %d: ", func, file, line);

	switch (status) {
		case UMFPACK_ERROR_out_of_memory:
			fprintf(stderr, "out of memory!\n");
			break;
		case UMFPACK_WARNING_singular_matrix:
			fprintf(stderr, "matrix is singular!\n");
			break;
		default:
			fprintf(stderr, "UMFPACK error code %d\n", status);
	}
}


void complex_solver_dump_matrix(int col,int nz,int *Ti,int *Tj, double *Tx, double *Txz,double *b,double *bz)
{
FILE *matrix;
matrix=fopen("./matrix.dat","w");


int i;
for (i=0;i<nz;i++)
{
	fprintf(matrix,"%d %d %le\n",Tj[i],Ti[i],pow(pow(Tx[i],2.0)+pow(Txz[i],2.0),2.0));
}

for (i=0;i<col;i++)
{
	fprintf(matrix,"%d %d %le\n",0,i,fabs(pow(b[i],2.0)+pow(bz[i],2.0)));
}

printf("Matrix dumped\n");

fclose(matrix);
}

void complex_solver_free()
{
free(x);
free(xz);
free(Ap);
free(Ai);
free(Ax);
free(Az);
x=NULL;
xz=NULL;
Ap=NULL;
Ai=NULL;
Ax=NULL;
Az=NULL;
last_col=0;
last_nz=0;
}

void complex_solver_print(int col,int nz,int *Ti,int *Tj, double *Tx, double *Txz,double *b,double *bz)
{
int i;
for (i=0;i<nz;i++)
{
	printf("%d %d %le+i%le\n",Ti[i],Tj[i],Tx[i],Txz[i]);
}

for (i=0;i<col;i++)
{
	printf("%le+i%le\n",b[i],bz[i]);
}


}

int complex_solver(int col,int nz,int *Ti,int *Tj, double *Tx, double *Txz,double *b,double *bz)
{

//getchar();
int i;
void *Symbolic, *Numeric;
int status;
//printf("here1\n");
if ((last_col!=col)||(last_nz!=nz))
{
x = realloc(x,col*sizeof(double));
xz = realloc(xz,col*sizeof(double));
Ap = realloc(Ap,(col+1)*sizeof(int));
Ai = realloc(Ai,(nz)*sizeof(int));
Ax  = realloc(Ax,(nz)*sizeof(double));
Az = realloc (Az,(nz) * sizeof (double));
last_col=col;
last_nz=nz;
}

double Info [UMFPACK_INFO], Control [UMFPACK_CONTROL];

/* get the default control parameters */
umfpack_zi_defaults (Control) ;

/* change the default print level for this demo */
/* (otherwise, nothing will print) */
Control [UMFPACK_PRL] = 1 ;

/* print the license agreement */
//umfpack_zi_report_status (Control, UMFPACK_OK) ;
Control [UMFPACK_PRL] = 0 ;

/* print the control parameters */
umfpack_zi_report_control (Control) ;

//status = umfpack_di_triplet_to_col(col, col, nz, Ti, Tj, Tx, Ap, Ai, Ax, NULL);
status = umfpack_zi_triplet_to_col (col, col, nz, Ti, Tj, Tx, Txz, Ap, Ai, Ax, Az, NULL) ;


if (status != UMFPACK_OK) {
	error_report(status, __FILE__, __func__, __LINE__);
	return EXIT_FAILURE;
}

/* symbolic analysis */
//printf("here2 %d\n",col);
//status = umfpack_di_symbolic(col, col, Ap, Ai, Ax, &Symbolic, NULL, NULL);
status = umfpack_zi_symbolic(col, col, Ap, Ai, Ax, Az, &Symbolic, Control, Info) ;
umfpack_zi_report_status (Control, status) ;
//printf("here3\n");

if (status != UMFPACK_OK) {
	error_report(status, __FILE__, __func__, __LINE__);
	return EXIT_FAILURE;
}

/* LU factorization */
//umfpack_di_numeric(Ap, Ai, Ax, Symbolic, &Numeric, NULL, NULL);
umfpack_zi_numeric (Ap, Ai, Ax, Az, Symbolic, &Numeric, Control, Info) ;

if (status != UMFPACK_OK) {
	error_report(status, __FILE__, __func__, __LINE__);
	return EXIT_FAILURE;
}
/* solve system */

//umfpack_di_free_symbolic(&Symbolic);

        // umfpack_di_solve(UMFPACK_A, Ap, Ai, Ax, x, b, Numeric, NULL, NULL);
status = umfpack_zi_solve(UMFPACK_A, Ap, Ai, Ax, Az, x, xz, b, bz, Numeric, Control, Info) ;
if (status != UMFPACK_OK) {
	error_report(status, __FILE__, __func__, __LINE__);
	return EXIT_FAILURE;
}

//printf ("\nx (solution of Ax=b): ") ;
    (void) umfpack_zi_report_vector (col, x, xz, Control) ;

umfpack_zi_free_symbolic (&Symbolic) ;
umfpack_di_free_numeric(&Numeric);

for (i = 0; i < col; i++)
{
	b[i]=x[i];
	bz[i]=xz[i];
}

return 0;
}
